/* $NetBSD$ */

/*-
 * Copyright (c)2008, 2011 Citrus Project,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
#if defined(LIBC_SCCS) && !defined(lint)
__RCSID("$NetBSD$");
#endif /* LIBC_SCCS and not lint */

#include <assert.h>
#include <errno.h>
#include <string.h>

#include "xlocale.h"
#include "setlocale_local.h"

#include "runetype_local.h"
#include "_wctype_local.h"
#include "_wctrans_local.h"

#define _RUNE_LOCALE(locale) ((_RuneLocale const *) \
    (_locale_to_impl(locale))->part_impl[(size_t)LC_CTYPE])

#define _RUNE_ISWCTYPE_FUNC(name, index)		\
int							\
_isw##name##_l(wint_t wc, _locale_t locale)		\
{							\
	_RuneLocale const *rl;				\
	_WCTypeEntry const *te;				\
							\
	rl = _RUNE_LOCALE(locale);			\
	te = &rl->rl_wctype[index];			\
	return _iswctype_priv(rl, wc, te);		\
}
_RUNE_ISWCTYPE_FUNC(alnum,  _WCTYPE_INDEX_ALNUM)
_RUNE_ISWCTYPE_FUNC(alpha,  _WCTYPE_INDEX_ALPHA)
_RUNE_ISWCTYPE_FUNC(blank,  _WCTYPE_INDEX_BLANK)
_RUNE_ISWCTYPE_FUNC(cntrl,  _WCTYPE_INDEX_CNTRL)
_RUNE_ISWCTYPE_FUNC(digit,  _WCTYPE_INDEX_DIGIT)
_RUNE_ISWCTYPE_FUNC(graph,  _WCTYPE_INDEX_GRAPH)
_RUNE_ISWCTYPE_FUNC(lower,  _WCTYPE_INDEX_LOWER)
_RUNE_ISWCTYPE_FUNC(print,  _WCTYPE_INDEX_PRINT)
_RUNE_ISWCTYPE_FUNC(punct,  _WCTYPE_INDEX_PUNCT)
_RUNE_ISWCTYPE_FUNC(space,  _WCTYPE_INDEX_SPACE)
_RUNE_ISWCTYPE_FUNC(upper,  _WCTYPE_INDEX_UPPER)
_RUNE_ISWCTYPE_FUNC(xdigit, _WCTYPE_INDEX_XDIGIT)

#define _RUNE_TOWCTRANS_FUNC(name, index)		\
wint_t							\
_tow##name##_l(wint_t wc, _locale_t locale)		\
{							\
	_RuneLocale const *rl;				\
	_WCTransEntry const *te;			\
							\
	rl = _RUNE_LOCALE(locale);			\
	te = &rl->rl_wctrans[index];			\
	return _towctrans_priv(wc, te);			\
}
_RUNE_TOWCTRANS_FUNC(upper, _WCTRANS_INDEX_UPPER)
_RUNE_TOWCTRANS_FUNC(lower, _WCTRANS_INDEX_LOWER)

wctype_t
_wctype_l(const char *charclass, _locale_t locale)
{
	_RuneLocale const *rl;
	size_t i;

	rl = _RUNE_LOCALE(locale);
	for (i = 0; i < _WCTYPE_NINDEXES; ++i) {
		if (!strcmp(rl->rl_wctype[i].te_name, charclass))
			return (wctype_t)&rl->rl_wctype[i];
	}
	return (wctype_t)NULL;
}

wctrans_t
_wctrans_l(const char *charmap, _locale_t locale)
{
	_RuneLocale const *rl;
	size_t i;

	rl = _RUNE_LOCALE(locale);
	for (i = 0; i < _WCTRANS_NINDEXES; ++i) {
		_DIAGASSERT(rl->rl_wctrans[i].te_name != NULL);
		if (!strcmp(rl->rl_wctrans[i].te_name, charmap))
			return (wctrans_t)&rl->rl_wctype[i];
	}
	return (wctrans_t)NULL;
}

int
_iswctype_l(wint_t wc, wctype_t charclass, _locale_t locale)
{
	_RuneLocale const *rl;
	_WCTypeEntry const *te;

	_DIAGASSERT(locale != NULL);

	if (charclass == NULL) {
		errno = EINVAL;
		return 0;
	}
	rl = _RUNE_LOCALE(locale);
	te = (_WCTypeEntry const *)charclass;
	return _iswctype_priv(rl, wc, te);
}

wint_t
_towctrans_l(wint_t wc, wctrans_t charmap, _locale_t locale)
{
	_RuneLocale const *rl;
	_WCTransEntry const *te;

	_DIAGASSERT(locale != NULL);

	if (charmap == NULL) {
		errno = EINVAL;
		return wc;
	}
	rl = _RUNE_LOCALE(locale);
	te = (_WCTransEntry const *)charmap;
	return _towctrans_priv(wc, te);
}

int
_wcwidth_l(wchar_t wc, _locale_t locale)
{
	_RuneLocale const *rl;
	_RuneType x;

	if (wc == L'\0')
		return 0;
	rl = _RUNE_LOCALE(locale);
	x = _runetype_priv(rl, wc);
	if (x & _RUNETYPE_R)
		return ((unsigned)x & _RUNETYPE_SWM) >> _RUNETYPE_SWS;
	return -1;
}

int
_wcswidth_l(const wchar_t * __restrict ws, size_t wn, _locale_t locale)
{
	_RuneLocale const *rl;
	_RuneType x;
	int width;

	_DIAGASSERT(ws != NULL);

	rl = _RUNE_LOCALE(locale);
	for (width = 0; wn > 0 && *ws != L'\0'; ++ws, --wn) {
		x = _runetype_priv(rl, *ws);
		if ((x & _RUNETYPE_R) == 0)
			return -1;
		width += ((unsigned)x & _RUNETYPE_SWM) >> _RUNETYPE_SWS;
	}
	return width;
}
